/*
** Module   : PortTypeWrapper.java
** Abstract : container that implements a port-type procedure.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 002 CA  20131220 Created initial version.
** 003 VIG  20140116 Changed currentWindow method signature.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.ProcedureManager.SearchMode;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * Implements a procedure connected to a WSDL port-type. Compared to the 
 * {@link ExternalProgramWrapper} implementation, port-type procedures are limited, most of the
 * APIs being no-op or returning default values. Only the following procedure-specific APIs are
 * implemented by port-types too:
 * <ul>
 *   <li>FILE-NAME</li>
 *   <li>PERSISTENT</li>
 *   <li>SERVER</li>
 *   <li>ASYNC-REQUEST-COUNT</li>
 *   <li>SET-CALLBACK-PROCEDURE</li>
 *   <li>SERVER-HANDLE</li>
 *   <li>CURRENT-WINDOW</li>
 * </ul>
 */
public class PortTypeWrapper
extends HandleChain
implements PersistentProcedure,
           AsyncRequestListener
{
   /** Constant identifying the <code>REQUEST-HEADER</code> callback. */
   private static final String CALLBACK_REQUEST_HEADER = "REQUEST-HEADER";
   
   /** Constant identifying the <code>RESPONSE-HEADER</code> callback. */
   private static final String CALLBACK_RESPONSE_HEADER = "RESPONSE-HEADER";
   
   /** The {@link ServerImpl} instance associated with this port-type procedure. */
   private final ServerImpl server;
   
   /** The ADM-DATA of this port-type procedure. */
   private final character admData = new character();
   
   /** The UNIQUE-ID of this port-type procedure. */
   private final integer uniqueId = new integer();
   
   /**
    * The name of request-header callback procedure.  Unknown if request-header callback is not 
    * used.
    */
   private final character requestHeaderProc = new character();

   /**
    * The context for the request-header procedure.  Unknown if request-header callback is not 
    * used.
    */
   private final handle requestHeaderContext = new handle();
   
   /**
    * The name of response-header callback procedure.  Unknown if response-header callback is not 
    * used.
    */
   private final character responseHeaderProc = new character();
   
   /**
    * The context for the response-header procedure.  Unknown if response-header callback is not 
    * used.
    */
   private final handle responseHeaderContext = new handle();
   
   /** The number of active asynchronous requests. */
   private long asyncRequestCount = 0;

   /** Flag indicating if this port-type procedure was deleted or not. */
   private boolean deleted = false;

   /**
    * Instantiate a new port-type procedure by specifying the port-type name and the associated
    * {@link ServerImpl} instance.
    * 
    * @param    name
    *           The port-type name.  This will be reported by the NAME and FILE-NAME attributes.
    * @param    server
    *           The associated server instance.
    */
   PortTypeWrapper(String name, ServerImpl server)
   {
      // port-type procedures do not collide with external or proxy procedures (thus the first/last
      // resource must not be set).
      super(false, false);

      this.name.assign(name);
      this.server = server;
      
      this.uniqueId.assign(UniqueIdGenerator.getUniqueId(IdKind.PROCEDURE));
   }

   /**
    * Reports if this object is valid for use.  
    * <p>
    * A port-type is valid only if it was not yet deleted.
    * 
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Get the <code>file-name</code> attribute of this handle.
    * 
    * @return   See above.
    */
   @Override
   public character getFileName()
   {
      return new character(name());
   }

   /**
    * Returns a handle to the app server in which the procedure is ran.
    * 
    * @return   See above
    */
   @Override
   public handle getServerHandle()
   {
      return new handle(server);
   }

   /**
    * Returns the number of asynchronous requests.
    * 
    * @return   See above.
    */
   @Override
   public integer getAsyncRequestCount()
   {
      return new integer(asyncRequestCount);
   }

   /**
    * Get the <code>persistent</code> attribute of this port-type procedure.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public logical isPersistent()
   {
      return new logical(true);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, String proc)
   {
      return setCallbackProcedure(new character(event), new character(proc),
                                  ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, String proc)
   {
      return setCallbackProcedure(event, new character(proc), ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, character proc)
   {
      return setCallbackProcedure(new character(event), proc, ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, character proc)
   {
      return setCallbackProcedure(event, proc, ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, String proc, handle context)
   {
      return setCallbackProcedure(new character(event), new character(proc), context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, String proc, handle context)
   {
      return setCallbackProcedure(event, new character(proc), context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, character proc, handle context)
   {
      return setCallbackProcedure(new character(event), proc, context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, character proc, handle context)
   {
      String evt = (event.isUnknown() ? "" : event.toStringMessage());
      
      if (!(evt.equalsIgnoreCase(CALLBACK_REQUEST_HEADER) ||
            evt.equalsIgnoreCase(CALLBACK_RESPONSE_HEADER)))
      {
         final int[] errCodes = { 11576, 4065 };
         final String msg1 = "Invalid callback type passed to SET-CALLBACK-PROCEDURE; %s. " +
            "Valid callback types are: REQUEST-HEADER or RESPONSE-HEADER (11576)";
         final String msg2 = "**The SET-CALLBACK-PROCEDURE attribute on the PROCEDURE widget " +
            "has invalid arguments. (4065)";
         
         ErrorManager.recordOrShowError(errCodes, new String[] {String.format(msg1, evt), msg2},
                                        false, false, false, true);

         return new logical(false);
      }
      
      if (!context._isValid() || !(context.getResource() instanceof PersistentProcedure))
      {
         final int[] errCodes = {11982, 4065};
         final String msg1 = 
            "Third argument to SET-CALLBACK-PROCEDURE must be a valid procedure handle";
         final String msg2 =
           "**The SET-CALLBACK-PROCEDURE attribute on the PROCEDURE widget has invalid arguments";
         ErrorManager.recordOrShowError(errCodes, new String[] {msg1, msg2},
                                        false, false, false);
         
         return new logical(false);
      }
      
      // save the procedure and context
      if (CALLBACK_REQUEST_HEADER.equalsIgnoreCase(evt))
      {
         if (proc.toStringMessage().length() == 0)
         {
            // reset them
            this.requestHeaderContext.setUnknown();
            this.requestHeaderProc.setUnknown();
         }
         else
         {
            this.requestHeaderContext.assign(context);
            this.requestHeaderProc.assign(proc.toStringMessage());
         }
      }
      else
      {
         if (proc.toStringMessage().length() == 0)
         {
            // reset them
            this.responseHeaderContext.setUnknown();
            this.responseHeaderProc.setUnknown();
         }
         else
         {
            this.responseHeaderContext.assign(context);
            this.responseHeaderProc.assign(proc.toStringMessage());
         }
      }
      
      // false is always returned
      return new logical(false);
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   See above.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueId);
   }

   /**
    * Implements the <code>PUBLISHED-EVENTS</code> attribute.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always the empty string.
    */
   @Override
   public character getPublishedEvents()
   {
      return new character("");
   }

   /**
    * Implements the <code>SUPER-PROCEDURES</code> attribute.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always the empty string.
    */
   @Override
   public character superProcedures()
   {
      return new character("");
   }

   /**
    * Implements the <code>ADD-SUPER-PROCEDURE</code> method.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical addSuperProcedure(handle h)
   {
      return new logical(false);
   }

   /**
    * Implements the <code>ADD-SUPER-PROCEDURE</code> method.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical addSuperProcedure(handle h, SearchMode mode)
   {
      return new logical(false);
   }

   /**
    * Implements the <code>ADD-SUPER-PROCEDURE</code> method.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical addSuperProcedure(handle h, int mode)
   {
      return new logical(false);
   }

   /**
    * Implements the <code>ADD-SUPER-PROCEDURE</code> method.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical addSuperProcedure(handle h, integer mode)
   {
      return new logical(false);
   }

   /**
    * Implements the <code>REMOVE-SUPER-PROCEDURE</code> method.
    * <p>
    * This is a no-op for port-type procedures.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical removeSuperProcedure(handle h)
   {
      return new logical(false);
   }

   /**
    * Implements the <code>REMOTE</code> attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical isRemote()
   {
      return new logical(false);
   }

   /**
    * Implements the <code>INTERNAL-ENTRIES</code> attribute.
    * 
    * @return   Always unknown.
    */
   @Override
   public character internalEntries()
   {
      return new character();
   }

   /**
    * Implements the <code>GET-SIGNATURE</code> method.
    * 
    * @return   Always the empty string.
    */
   @Override
   public character getSignature(String internalEntry)
   {
      return new character("");
   }

   /**
    * Implements the <code>GET-SIGNATURE</code> method.
    * 
    * @return   Always the empty string.
    */
   @Override
   public character getSignature(character internalEntry)
   {
      return new character("");
   }

   /**
    * Get CURRENT-WINDOW attribute for given procedure handle.
    * <p>
    * 
    * @return   Alwasy unknown.
    */
   @Override
   public handle currentWindow()
   {
      // TODO: return unknown
      return null;
   }

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * <p>
    * This is a no-op, but the handle is validated.
    * 
    * @param    h
    *           A handle for the window.
    */
   @Override
   public void setCurrentWindow(handle h)
   {
      // TODO: validate it
   }

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * <p>
    * This is a no-op, but the window is validated.
    * 
    * @param    win
    *           Window reference.
    */
   @Override
   public void setCurrentWindow(WindowWidget win)
   {
      // TODO: validate it
   }

   /**
    * This is a no-op, always return <code>false</code>.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public logical isProxy()
   {
      return new logical(false);
   }

   /**
    * This is a no-op, always return unknown handle.
    * 
    * @return   Always unknown handle.
    */
   @Override
   public handle getTransaction()
   {
      return new handle();
   }
   
   /**
    * Called when an internal-entry of this proxy procedure is invoked async.
    */
   @Override
   public synchronized void notifyStart()
   {
      asyncRequestCount++;
   }

   /**
    * Called when the async call of an internal-entry for proxy procedure has finished.
    */
   @Override
   public synchronized void notifyFinish()
   {
      asyncRequestCount--;
   }

   /**
    * Check if this port-type procedure can be deleted.
    * 
    * @return   <code>true</code> if the port-type procedure can be deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (asyncRequestCount > 0)
      {
         // can't delete if async req count > 0
         final String msg = "Handle value supplied to the DELETE PROCEDURE/OBJECT statement " +
            "has outstanding asynchronous requests. Procedure '%s'";

         ErrorManager.recordOrThrowError(8980, String.format(msg, name.toStringMessage()));
         return false;
      }
      
      if (!server.valid())
      {
         final String err =
            "Invalid or inappropriate SERVER handle implied by DELETE PROCEDURE statement";
         ErrorManager.displayError(5455, err, false);
         return false;
      }

      if (!server._connected())
      {
         final String err = "SERVER  is not connected; cannot delete procedure handle or object";
         ErrorManager.displayError(5456, err, false);
         return false;
      }

      deleted = true;
      return true;
   }
   
   /**
    * Get the {@link ServerImpl} instance associated with this port-type.
    * 
    * @return   See above.
    */
   ServerImpl getServer()
   {
      return this.server;
   }

   /**
    * Get the {@link #requestHeaderProc} for this port-type. When <code>null</code>, then no
    * <code>request-header</code> procedure is set.
    * 
    * @return   See above.
    */
   String getRequestHeaderProc()
   {
      return requestHeaderProc.isUnknown() ? null : requestHeaderProc.toStringMessage();
   }
   
   /**
    * Get the {@link #requestHeaderContext} for this port-type.
    * 
    * @return   See above.
    * 
    * @see   #getRequestHeaderProc()
    */
   handle getRequestHeaderContext()
   {
      return requestHeaderContext;
   }
   
   /**
    * Get the {@link #responseHeaderProc} for this port-type. When <code>null</code>, then no
    * <code>response-header</code> procedure is set.
    * 
    * @return   See above.
    */
   String getResponseHeaderProc()
   {
      return responseHeaderProc.isUnknown() ? null : responseHeaderProc.toStringMessage();
   }
   
   /**
    * Get the {@link #responseHeaderContext} for this port-type.
    * 
    * @return   See above.
    * 
    * @see   #getResponseHeaderProc()
    */
   handle getResponseHeaderContext()
   {
      return responseHeaderContext;
   }
}
